
\ Callist2.f 
\ Каллисто-2, черновик 6
\ Авторское право (c) Васильев Илья Владимирович, Москва, 15 ноября 2016
\ Каллисто-2 это язык программирования для "Электроники МК-161", развивающий идеи Каллисто Классик

\ Каллисто-2 распространяется под свободной лицензией GNU GPL v3 или выше
\ Если вы несогласны с правами, которые предоставляет свободное программное обеспечение
\ и обязанностями, которые оно налагает, удалите Каллисто-2 со своих носителей.

\ Исходный текст предназначен для SP-Forth 4.20

\ == Регистры МК-161 ==
\ R2 RP        Указатель целочисленного стека возвратов, растёт вниз от 5090
\ R5           Временный регистр обработчика, может быть изменён NEXT, CALL, EXIT
\ R6 RI, IP    Указатель инструкций, указывает на токен в шитом коде
\ R7 W, WP     Указатель слов Форта, содержит CFA исполняемого слова
\              временный регистр обработчика
\ R8 JP        Указатель/регистр передачи управления NEXT,
\              временный регистр обработчика
\ R9 =NEXT     Адрес точки NEXT
\ RC TOS       Вершина стека данных
\
\ R9042 =9500  Индексный регистр памяти программ
\              Содержит адрес таблицы свёрток

ONLY FORTH DEFINITIONS
CREATE prgm  10000 ALLOT         \ Память программ
prgm 10000 255 FILL              \ Заполняем FF

CREATE regb  8168 ALLOT          \ Регистровая память (байты)
regb 8168 ERASE                  \ Обнуляем

: save.mkp ( -- )                \ Сохранить память программ в файл Callist2.mkp
  HERE 0 C,                      \ Номер начальной страницы
  prgm 10000 0 DO DUP C@ C, 1 C, 1+ LOOP DROP
  S" Callist2.mkp" W/O ( BIN )
  CREATE-FILE ABORT" Can't create MKP"
  OVER 20001 2 PICK
  WRITE-FILE  ABORT" Can't write MKP"
  CLOSE-FILE  ABORT" Can't close MKP"
  -20001 ALLOT
;

: save.mkb ( -- )                \ Сохранить байтовые регистры в файл Callist2.mkb
  S" Callist2.mkb" W/O ( BIN )
  CREATE-FILE ABORT" Can't create MKB"
  1000 regb + 4096 2 PICK
  WRITE-FILE  ABORT" Can't write MKB"
  CLOSE-FILE  ABORT" Can't close MKB"
;

\ --------------------------------------------------------
\ X-компилятор исходного текста Каллисто-2 в память МК-161
\ --------------------------------------------------------

: 16* ( n1 -- n2 )  2* 2* 2* 2* ;                   \ Умножение на 16
: [+] + ;  : [-] - ;                                \ Сохраним операции со стеком Форта, как [+] и [-]
: getnumber ( "aaaa" -- )                           \ Считать следующее слово из входного потока, как число
    00 00 BL WORD COUNT >NUMBER 2DROP DROP ;

1000 VALUE bc
: bhere ( -- a ) bc regb [+] ;                      \ Текущий байтовый регистр
: ballot ( n -- ) bc [+] TO bc ;                    \ Зарезервировать n байтовых регистров
: b, ( c -- ) bhere C! 01 ballot ;                  \ Скомпилировать байт c
: bxt, ( xt -- ) 100 /MOD b, b, ;                   \ Скомпилировать выполнимый токен xt
: .borg ( "aaaa" -- ) getnumber TO bc ;             \ Сделать байтовый регистр aaaa текущим

00 VALUE pc                                         \ Счётчик команд
00 VALUE sc                                         \ Номер свободной свёртки
: phere ( -- a )  pc prgm [+] ;                     \ Адрес текущего места в памяти программ
: pallot ( n -- )  pc [+] TO pc ;                   \ Зарезервировать n шагов в памяти программ
: p, ( c -- )  phere C! 01 pallot ;                 \ Скомпилировать шаг c в память программ
: pbcd1, ( n -- ) 100 MOD 10 /MOD 16* [+] p, ;      \ Скомпилировать в шаг младшую половину n, как число bcd
: pbcd2, ( n -- ) 100 /MOD pbcd1, pbcd1, ;          \ Скомпилировать в два шага bcd-число n
: ps, ( pc -- ) 9500 prgm [+]                       \ Добавить адрес обработчика pc в таблицу свёрток
   sc 2* [+]  SWAP 256 /MOD
   2 PICK C!  SWAP 1+ C!  sc 1+ TO sc ;
: A: ( "имя" -- )  CREATE pc , DOES> ( -- pc ) @ ;  \ Метка
: .org ( "aaaa" -- )  getnumber TO pc ;             \ Сделать шаг по адресу aaaaa текущим



\ ------------------------------------------------
\ X-компилятор входного языка "Электроники МК-161"
\ ------------------------------------------------

: cop ( cop -- )  CREATE C, DOES> ( -- ) C@ p, ;                 \ Одношаговая команда
: 2op ( cop -- )  CREATE C, DOES> ( "xx" -- ) C@ p,  BASE @ HEX  \ Двухшаговая команда
    getnumber p,  BASE ! ;
: 3op ( cop -- )  CREATE C, DOES> ( "xx" -- ) C@ p,  BASE @ HEX  \ Трёхшаговая команда
    getnumber 256 /MOD p, p,  BASE ! ;

VOCABULARY ЯМК

: code ( "имя" -- )                                 \ Начать x-определение примитива
    CREATE bc ,  sc b,  pc ps,  ALSO ЯМК  DOES> ( -- ) @ bxt, ;
: end-code ( -- ) ONLY FORTH ;                      \ Завершить x-определение примитива

: :: ( "имя" -- )                                   \ Начать x-определение высокого уровня
    CREATE bc ,  00 b,  DOES> ( -- ) @ bxt, ;
: ;; ( -- )  ;                                      \ Может компилировать exit, но пока пишем exit вручную

: ?pairs ( n1 n2 -- )                               \ Всё ли в порядке с операторными скобками?
    - ABORT" Conditionals aren't paired" ;
: ?far ( цель -- флаг )                             \ Проверить необходимость межстраничного перехода
    100 /  pc 1+ 100 /  = 0= ;

ONLY FORTH ALSO ЯМК DEFINITIONS
 07 CONSTANT x≠0                                    \ четыре условия МК-161
 09 CONSTANT x≥0
0xC CONSTANT x<0
0xE CONSTANT x=0

: begin ( -- цель 101 )  pc 101 ; IMMEDIATE         \ Отметить, куда будет переход назад, начало цикла
: until ( цель 101 условие -- ) SWAP 101 ?pairs     \ Условный переход назад, цикл "до"
    OVER ?far IF 0xF0 [+] p, pbcd2, ELSE 0x50 [+] p, pbcd1, THEN ; IMMEDIATE
: again ( цель 101 -- ) 101 ?pairs                  \ Безусловный переход назад, бесконечный цикл
    DUP ?far IF 0xF1 p, pbcd2, ELSE 0x51 p, pbcd1, THEN ; IMMEDIATE
: pnum, ( число -- ) -1 SWAP                        \ Скомпилировать целое число в команды языка МК-161
    BEGIN 10 /MOD DUP 0= UNTIL DROP                 \ Сперва кладём все цифры в стек
    BEGIN p, DUP 0< UNTIL DROP                      \ Потом компилируем их в нужном порядке
;

   0 cop 0        1 cop 1          2 cop 2          3 cop 3         4 cop 4        5 cop 5        6 cop 6        7 cop 7
   8 cop 8        9 cop 9        0xA cop ,        0xB cop /-/     0xC cop ВП     0xD cop Cx     0xE cop В↑     0xF cop FВх
0x10 cop +     0x11 cop -       0x12 cop ×       0x13 cop ÷      0x14 cop ↔     0x15 cop F10ˣ  0x16 cop Feˣ   0x17 cop Flg
0x18 cop Fln   0x19 cop Farcsin 0x1A cop Farccos 0x1B cop Farctg 0x1C cop Fsin  0x1D cop Fcos  0x1E cop Ftg
0x20 cop Fπ    0x21 cop F√      0x22 cop Fx²     0x23 cop F1/x   0x24 cop Fxʸ   0x25 cop Fo    0x26 cop КМ→Г  0x27 cop K-
0x28 cop КИПРГ 0x29 cop К÷      0x2A cop КМС→Г
0x30 cop КГ→МС 0x31 cop К|x|    0x32 cop КЗН     0x33 cop КГ→М   0x34 cop К[x]  0x35 cop К{x}  0x36 cop Кmax  0x37 cop КAND
0x38 cop КOR   0x39 cop КXOR    0x3A cop КNOT    0x3B cop КСЧ
0x40 cop П0    0x41 cop П1      0x42 cop П2      0x43 cop П3     0x44 cop П4    0x45 cop П5    0x46 cop П6    0x47 cop П7
0x48 cop П8    0x49 cop П9      0x4A cop ПА      0x4B cop ПВ     0x4C cop ПС    0x4D cop ПД    0x4E cop ПЕ    0x4F 2op РП
0x50 cop С/П   0x51 2op БП      0x52 cop В/О     0x53 2op ПП     0x54 cop КНОП  0x55 cop КЭКР  0x56 cop КГРФ  0x57 2op Fx≠0

0x60 cop ИП0   0x61 cop ИП1     0x62 cop ИП2     0x63 cop ИП3    0x64 cop ИП4   0x65 cop ИП5   0x66 cop ИП6   0x67 cop ИП7
0x68 cop ИП8   0x69 cop ИП9     0x6A cop ИПА     0x6B cop ИПВ    0x6C cop ИПС   0x6D cop ИПД   0x6E cop ИПЕ   0x6F 2op РИП
0x70 cop Кx≠00 0x71 cop Кx≠01   0x72 cop Кx≠02   0x73 cop Кx≠03  0x74 cop Кx≠04 0x75 cop Кx≠05 0x76 cop Кx≠06 0x77 cop Кx≠07
0x78 cop Кx≠08 0x79 cop Кx≠09   0x7A cop Кx≠0А   0x7B cop Кx≠0В  0x7C cop Кx≠0С 0x7D cop Кx≠0Д 0x7E cop Кx≠0Е 0x7F 2op РКx≠0
0x80 cop КБП0  0x81 cop КБП1    0x82 cop КБП2    0x83 cop КБП3   0x84 cop КБП4  0x85 cop КБП5  0x86 cop КБП6  0x87 cop КБП7
0x88 cop КБП8  0x89 cop КБП9    0x8A cop КБПА    0x8B cop КБПВ   0x8C cop КБПС  0x8D cop КБПД  0x8E cop КБПЕ  0x8F 2op РКБП
0x90 cop Кx≥00 0x91 cop Кx≥01   0x92 cop Кx≥02   0x93 cop Кx≥03  0x94 cop Кx≥04 0x95 cop Кx≥05 0x96 cop Кx≥06 0x97 cop Кx≥07
0x98 cop Кx≥08 0x99 cop Кx≥09   0x9A cop Кx≥0А   0x9B cop Кx≥0В  0x9C cop Кx≥0С 0x9D cop Кx≥0Д 0x9E cop Кx≥0Е 0x9F 2op РКx≥0
0xA0 cop КПП0  0xA1 cop КПП1    0xA2 cop КПП2    0xA3 cop КПП3   0xA4 cop КПП4  0xA5 cop КПП5  0xA6 cop КПП6  0xA7 cop КПП7
0xA8 cop КПП8  0xA9 cop КПП9    0xAA cop КППА    0xAB cop КППВ   0xAC cop КППС  0xAD cop КППД  0xAE cop КППЕ  0xAF 2op РКПП
0xB0 cop КП0   0xB1 cop КП1     0xB2 cop КП2     0xB3 cop КП3    0xB4 cop КП4   0xB5 cop КП5   0xB6 cop КП6   0xB7 cop КП7
0xB8 cop КП8   0xB9 cop КП9     0xBA cop КПА     0xBB cop КПВ    0xBC cop КПС   0xBD cop КПД   0xBE cop КПЕ   0xBF 2op РКП
0xC0 cop Кx<00 0xC1 cop Кx<01   0xC2 cop Кx<02   0xC3 cop Кx<03  0xC4 cop Кx<04 0xC5 cop Кx<05 0xC6 cop Кx<06 0xC7 cop Кx<07
0xC8 cop Кx<08 0xC9 cop Кx<09   0xCA cop Кx<0А   0xCB cop Кx<0В  0xCC cop Кx<0С 0xCD cop Кx<0Д 0xCE cop Кx<0Е 0xCF 2op РКx<0
0xD0 cop КИП0  0xD1 cop КИП1    0xD2 cop КИП2    0xD3 cop КИП3   0xD4 cop КИП4  0xD5 cop КИП5  0xD6 cop КИП6  0xD7 cop КИП7
0xD8 cop КИП8  0xD9 cop КИП9    0xDA cop КИПА    0xDB cop КИПВ   0xDC cop КИПС  0xDD cop КИПД  0xDE cop КИПЕ  0xDF 2op РКИП
0xE0 cop Кx=00 0xE1 cop Кx=01   0xE2 cop Кx=02   0xE3 cop Кx=03  0xE4 cop Кx=04 0xE5 cop Кx=05 0xE6 cop Кx=06 0xE7 cop Кx=07
0xE8 cop Кx=08 0xE9 cop Кx=09   0xEA cop Кx=0А   0xEB cop Кx=0В  0xEC cop Кx=0С 0xED cop Кx=0Д 0xEE cop Кx=0Е 0xEF 2op РКx=0

0xF2 cop РРВ/О
0xF4 3op РРП
0xF6 3op РРИП


\ ----------------------
\ Память программ МК-161
\ ----------------------

Cx                       \ === Инициализация ===
9 5 0 0 РРП 9042         \ Адрес таблицы свёрток
5 0 9 0 П2               \ Инициализация стека возвратов
1 0 9 9 П6               \ RI := Начало шитого кода - 1
pc 03 [+] pnum,          \ Здесь мы подразумевам, что адрес NEXT меньше 100
П9                       \ Занесём в R9 адрес NEXT


A: NEXT                  \ Адресный интерпретатор, выполнить следующий токен в шитом коде
КИП6 ВП 2 КИП6 + П7      \ W := REGW[RI++], считываем CFA в R7
КИП7 РРП 9210            \ Находим по CFA адрес обработчика
П8 КБП8                  \ Передаём на него управление

A: CALL                  \ Обработчик слов высокого уровня
ИП6 В↑ ВП 2 /-/ К[x] КП2 \ Сохраняем старшую половину RI
ВП 2 - КП2               \ Сохраняем младшую половину RI
ИП7 П6                   \ RI := PFA - 1 = CFA
КБП9                     \ Для быстродействия CALL или RETURN надо склеить с NEXT

CALL ps,                 \ Обработчик 000: CALL

ONLY FORTH DEFINITIONS

code exit             \ === Примитив exit ===
  ИП2 П8 П5  2 + П2   \ RP += 2
  КИП5 ВП 2 КИП8 + П6 \ RPOP RI
  КБП9                \ NEXT
end-code

code cls              \ === Примитив cls ===
  2 РРП 9010          \ Инициализация графического экрана
  КБП9                \ NEXT
end-code

code setA             \ === Примитив setA ===
  1 2 8 ПС            \ Записать в RC код буквы "А"
  КБП9                \ NEXT
end-code

code 1+               \ === Примитив 1+ ===
  ИПС 1 + ПС          \ Увеличить RC на 1
  КБП9                \ NEXT
end-code

code 1-               \ === Примитив 1- ===
  ИПС 1 - ПС          \ Уменьшить RC на 1
  КБП9                \ NEXT
end-code

code emit             \ === Примитив emit ===
  ИПС РРП 9020        \ Вывод символа из RC на графический экран
  КГРФ                \ Обновляем индикатор
  КБП9                \ NEXT
end-code

code key              \ === Примитив key ===
  begin               \ Начало цикла ожидания нажатия на клавишу
    РРИП 9029         \ Чтение кода нажатой клавиши
    КNOT              \ Проверка на 255
  x≠0 until           \ Дождёмся чего-нибудь, отличного от 255
  РРИП 9000           \ Программирование начальных координат объекта
  Cx 8 + 0 РРП 9000   \ Перейти в начало следующей строки
  КБП9                \ NEXT
end-code

code bye              \ === Примитив bye ===
  С/П                 \ Выход в Калькулятор
end-code

:: testAAA            \ Определения высокого уровня, пока для проверки
     setA emit emit emit
     exit
;;

:: testBBB
     setA 1+ emit testAAA
     setA 1+ emit testAAA
     setA 1+ emit
     exit
;;

:: testVVV
     setA 1+ 1+ emit testAAA
     setA 1+ 1+ emit testBBB
     setA 1+ 1+ emit
     key exit
;;


\ ------------------------
\ Байтовые регистры МК-161
\ ------------------------

.borg 1100        \ === Автозапускающийся шитый код ===
cls
setA 1+ emit key
testAAA key testBBB key testVVV bye


.( Compiled! )

\ Запишем Каллисто-2 на диск
save.mkp save.mkb

.( Saved!) CR

\EOF
